import os
import sys
import compiler
from parser import ParserError
from bike.parsing.load import load as loadast, getRootDirectory
from bike.refactor import extractMethod
from bike.refactor.rename import rename_new
from bike.refactor.extractMethod import coords
from bike.parsing.save import save as saveUpdates
from bike.parsing.utils import fqn_rcar, fqn_rcdr
from bike.parsing import visitor
from bike.parsing.undo import getUndoStack, UndoStackEmptyException
from bike.parsing.fastparserast import Class, Function
from bike.query.common import getScopeForLine
from bike.query.getTypeOf import getTypeOfExpr, UnfoundType
from bike.query.findReferences import findReferences
from bike.query.findDefinition import findAllPossibleDefinitionsByCoords
from bike.refactor import inlineVariable, extractVariable
from bike.parsing.newstuff import getRoot
from bike import log

def load(path):
    context = init()
    context.load(path)
    return context

def init():
    context = BRMContext_impl()
    return context

# the context object public interface
class BRMContext(object):


    def save(self):
        """ save the changed files out to disk """

    def setRenameMethodPromptCallback(self, callback):
        """
        sets a callback to ask the user about method refs which brm
        can't deduce the type of. The callback must be callable, and
         take the following parameters:
          - filename
          - linenumber
          - begin column
          - end column
         (begin and end columns enclose the problematic method call)
        """
        
    def renameByCoordinates(self, filename_path, line, col, newname):
        """ an ide friendly method which renames a class/fn/method
        pointed to by the coords and filename"""

    def extract(self, filename_path, 
                begin_line, begin_col,
                end_line, end_col, 
                name):
        """ extracts the region into the named method/function based on context"""

    def inlineLocalVariable(self,filename_path, line, col):
        """ Inlines the variable pointed to by
        line:col. (N.B. line:col can also point to a reference to the
        variable as well as the definition) """


    def extractLocalVariable(self,filename_path, begin_line, begin_col,
                             end_line, end_col, variablename):
        """ Extracts the region into a variable """

    # must be an object with a write method
    def setProgressLogger(self,logger):
        pass
        
    def undo(self):
        """ undoes the last refactoring. WARNING: this is dangerous if
        the user has modified files since the last refactoring.
        Raises UndoStackEmptyException"""

    def findReferencesByCoordinates(self, filename_path, line, column):
        """ given the coords of a function, class, method or variable
        returns a generator which finds references to it.
        """

    def findDefinitionByCoordinates(self,filename_path,line,col):        
        """ given the coordates to a reference, tries to find the
        definition of that reference """
        
class NotAPythonModuleOrPackageException: pass
class CouldntLocateASTNodeFromCoordinatesException: pass

class BRMContext_impl(BRMContext):
    
    def __init__(self):
        self.ast = getRoot()

        # Used because some refactorings delegate back to the user.
        # this flag ensures that code isnt imported during those times
        self.readyToLoadNewCode = 1 
        self.paths = []
        getUndoStack(1)  # force new undo stack
        self.promptUserClientCallback = None

    def load(self, path):
        if not self.readyToLoadNewCode:
            return
            
        if(not path.endswith(".py")) and os.path.isfile(path):
            raise NotAPythonModuleOrPackageException
        path = os.path.expanduser(path)
        path = os.path.abspath(path)
        self.paths.append(path)
        class SourceVisitor:
            def visitSource(self, node):
                if node.filename == path:
                    # backup the code
                    getUndoStack().addSource(node.filename,node.getSource())
                    getUndoStack().commitUndoFrame()
        if self.ast:
            compiler.walk(self.ast, SourceVisitor())
        self.ast = loadast(path, self.ast)


    def isLoaded(self, filename):
        if not filename.endswith(".py") or not os.path.isfile(filename):
            return 0
        if not self.ast:
            return 0
        filename = os.path.expanduser(filename)
        filename = os.path.abspath(filename)
        class SourceVisitor:
            def __init__(self):
                self.result = 0
            def visitSource(self, node):
                if node.filename == filename:
                    self.result = 1
        v = SourceVisitor()
        compiler.walk(self.ast, v)
        return v.result


    def _getAST(self):
        return self.ast

    # returns a list of saved filenames
    def save(self):
        savedfiles = saveUpdates()
        return savedfiles

    def setRenameMethodPromptCallback(self, callback):
        self.promptUserClientCallback = callback


    def extractMethod(self, filename_path, 
                        begin_line, begin_column, 
                        end_line, end_column, 
                        methodname):
        self.extract(filename_path, begin_line, begin_column, end_line, end_column,methodname)

    def extractFunction(self, filename_path, 
                        begin_line, begin_column, 
                        end_line, end_column, 
                        methodname):
        self.extract(filename_path, begin_line, begin_column, end_line, end_column,methodname)

    # does it based on context
    def extract(self, filename_path, 
                begin_line, begin_col,
                end_line, end_col, 
                name):
        self.refreshASTFromFileSystem()
        #ast = self._getSourceNode(filename_path)
        extractMethod.extractMethod(filename_path, coords(begin_line, begin_col), 
                                        coords(end_line, end_col), name)

    def inlineLocalVariable(self,filename_path, line, col):
        self.refreshASTFromFileSystem()
        ast = self._getSourceNode(filename_path)
        inlineVariable.inlineLocalVariable(ast,line,col)

    def extractLocalVariable(self,filename_path, begin_line, begin_col,
                             end_line, end_col, variablename):
        self.refreshASTFromFileSystem()
        ast = self._getSourceNode(filename_path)
        extractVariable.extractLocalVariable(ast,
                                          coords(begin_line, begin_col),
                                          coords(end_line, end_col),
                                          variablename)



    def undo(self):
        getUndoStack().undo()

    def _promptUser(self, filename, lineno, colbegin, colend):
        return self.promptUserClientCallback(filename, lineno, colbegin, colend)


    # must be an object with a write method
    def setProgressLogger(self,logger):
        log.progress = logger


    # filename_path must be absolute
    def renameByCoordinates(self, filename_path, line, col, newname):
        self.setNonLibPythonPath(filename_path)
        rename_new(filename_path,line,col,newname,
                   self.promptUserClientCallback)

    def _reverseCoordsIfWrongWayRound(self, colbegin, colend):
        if(colbegin > colend):
            colbegin,colend = colend,colbegin
        return colbegin,colend


    def findDefinitionByCoordinates(self,filename_path,line,col):
        self.setCompletePythonPath(filename_path)
        return findAllPossibleDefinitionsByCoords(filename_path,line,col)

        
    # filename_path must be absolute
    def findReferencesByCoordinates(self, filename_path, line, column):
        self.setNonLibPythonPath(filename_path)
        #print "pythonpath=",self.ast.pythonpath
        return findReferences(filename_path,line,column)
        

    def _getSourceNode(self, filename):
        return self.ast.getSourceNode(filename)

    def refreshASTFromFileSystem(self):
        for path in self.paths:
            self.ast = loadast(path, self.ast)


    def setCompletePythonPath(self,filename):
        rootdir = getRootDirectory(filename)
        pythonpath = [] + sys.path  # make a copy

        while rootdir in pythonpath:
            del pythonpath[pythonpath.index(rootdir)]
            
        self.ast.pythonpath = [rootdir] + pythonpath
        
    def setNonLibPythonPath(self,filename):
        if getRoot().unittestmode:
            return
        pythonpath = self.removeLibdirsFromPath(sys.path)
        pythonpath = [os.path.abspath(p) for p in pythonpath]
        rootdir = getRootDirectory(filename)
        while rootdir in pythonpath:
            del pythonpath[pythonpath.index(rootdir)]
        self.ast.pythonpath = [rootdir] + pythonpath

    def removeLibdirsFromPath(self, pythonpath):
        libdir = os.path.join(sys.prefix,"lib").lower()
        pythonpath = [p for p in pythonpath
                      if not p.lower().startswith(libdir)]
        return pythonpath

    #def removeDuplicates()
        
def _deducePackageOfFile(filename):
    package = ""
    dot = ""
    dir = os.path.dirname(filename)
    while dir != ""and \
          os.path.exists(os.path.join(dir, "__init__.py")):
        dir, dirname = os.path.split(dir)
        package = dirname+dot+package
        dot = "."
    return package
